#    The Circle - Decentralized resource discovery software
#    Copyright (C) 2001  Paul Francis Harrison
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""Circle application
"""

#256 color stuffed

#Still slows down on Windows

#Detect and use python2 if available? Or package for linux as well?

#Receive while senderer going offline: sender reappears!

# TODO:
#   - internationalization
#   French/German - njh, Japanese - sarahg

# BUG: pde & njh
#   - proxy keeps crashing

# TODO:
#   - option to publish apt cache in daemon

# TODO: ryan
# It's complicated telling someone how to find me on circle:
# Type "Peter Moulder" in the Search: field, click "Find 
# Person", click "View", then click "Remember this person between 
# sessions", click OK and close the search results window,
# then type "/join ryan" and hit return, then say hi.
# If I suggest that someone use circle to contact me and the instructions
# for doing so are so long (even without counting installing circle...
# Hmm, will have to set up an apt source for circle), then they'll probably
# not bother using circle.
# I'd guess that the main reason that one would Search for a Person would
# be to start talking to them, so how about streamlining this case?
# E.g. perhaps duplicate more of the View dialog into the person search results
# window, including a button that adds the person to acquaintance list and /join'sthem and probably dismisses the search window too.

# TODO: clausen
#   - status line (network usage, etc) to go in Network monitor window

# TODO: Brendan
#   - pageup/pagedown, search in gossip

# TODO: Ovind Stegard
#   - download speed in kb/s

# TODO: Kim Oldfield
#   - more type checking on incoming calls in node, name_server

# TODO: Thomas Mangin
#   - give daemons names

# TODO: pfh
#   - offline messages handled correctly when quiet

# TODO: mark
#   - automate copying of identity

# TODO: njh
#   - circle-person:dslkfjdsklfjsldkjfsalkj
#   - download whole directory

# TODO: Andrew Pam
#   - download speed, eta

# TODO: K Dobson
#   [modified win32 mainloop, fixed now?] - slow as buggery on WinXP
#   - uses lots of memory (leak?)

# TODO: insanity
#   - upload throttle

# TODO: Oliver Migeot
#   - ability to ignore people

# TODO: ryan
#   - documentation

# TODO: debbiep
#   - if in kde, use kde browser?
#   [obsoleted by channels?] - acquaintance classes for gossip
#circle's completion could never be as sophisticated as zsh's
#...
#it is just that zsh has very very powerful completion, much more
#than just filenames.
#I do appreciate that you can shell-escape from circle, but I'm
#much more likely to want to go the other way because 98% of the time
#I'm using the shell and don't want to have to type // in front of
#every command.
#I guess it's the polling vs interrupts argument all over again.
#circle's great in interrupt mode, but my work model requires
#polling.


# TODO: (sarahg's suggestions)
#   - multi window chat
#
#   - skins (for circle diagram, search area)
#   - configurable font size
#   - delete acquaintance
#   - no spaces in nicks

# TODO: njh
#   [obsoleted by /join + ?] - ability to use nick: to autospecify new recipient
#   *  whiteboard
#   * ability to hook into send and receive of chat messages
#   * faces
#   [obsoleted by channels?] - recent chat groups menu
#   !!!!!! remove excessive "from module import *"s

# TODO: jiri
#   - click on "Today's hot gossip" items to view
#   [ somewhat obsoleted by popularity sorting ] - trust network for mp3s
#   - retrieve offline messages only after have notified all watchers

# TODO: zaphod
#   - option to auto quiet after n minutes !! (maybe monitor mouse to see if person active...)
#   - resuming dowloads
#   - spell checker in gossip?
#
# TODO: bug1
#   - full path of files to be downloaded, for people who store artist/album/song.ogg

import os,os.path,sys,threading,time,traceback,md5,whrandom,re,string,math,types,imp,socket,signal

from settings import *
from utility import *
from node import *
from file_server import *
from cache import *
from chat import *
from channels import *
from name_server import *
from error import Error
import hash
import __init__

if use_gtk:
    from gtk import *
    import GDK #, GtkExtra

    from gossip import *
    from widgets import *
    from search import *
    import node_monitor

def absolutize_path():
    for i in range(len(sys.path)):
        sys.path[i] = os.path.abspath(sys.path[i])
     
def parse_address(local_address, str):
    try:
        match = re.match(r"\s*(.+)\s*:\s*([0-9]+)\s*$", str)
        if match:
            return (socket.gethostbyname(match.group(1)), string.atoi(match.group(2)))

        match = re.match(r"\s*([0-9]+)\s*$", str)
        if match:
            return (local_address[0], string.atoi(match.group(1)))
    
        return (socket.gethostbyname(string.strip(str)), default_ports[0])

    except:
        raise Error(_("Couldn't find peer:\n\n"+\
                                "Please give a valid IP address or machine name,\n"+\
                                "optionally followed by a colon and a port number"))

class Circle_application(Task_manager):
    """Coordinates the running of various services, and provides much of
         the GTK user interface."""

    config_keys = [
        ('name', 'username'),
        ('human-name', 'name'),
        ('description', 'finger')
    ]

    def __init__(self):
        Task_manager.__init__(self)
    
        self.config = get_config("configuration", { })
        for pair in [('root',''),('download',''),('name',os.environ.get('USER','anon')),('human-name',''),('description',''),('publish_apt_cache',0)]:
            if not self.config.has_key(pair[0]):
                self.config[pair[0]] = pair[1]

        self.config_dialog_open = 0
    
    def run_daemon(self):
        if not os.fork():
            os.setsid()
            if sys.platform <> 'win32':
                absolutize_path()
                os.chdir(os.environ['HOME'])
            sys.stdin.close()
            dev_null = open('/dev/null','wt')
            sys.stdout.close()
            sys.stderr.close()
            sys.stdout = dev_null
            sys.stderr = dev_null

            com_sock = socket.socket(socket.AF_UNIX,socket.SOCK_STREAM)
            com_sock_file = os.path.join(config_dir,'daemon_socket')

            try:
                os.unlink(com_sock_file) 
            except OSError:
                pass
            
            com_sock.bind(com_sock_file)
            com_sock.listen(5)
        
            self.node = Node()
            self.file_server = File_server(self.node)
            self.cache = Cache(self.node)

            self.start()
        
            self.node.start()

            pub_dir = get_config('daemon_public','')
            if pub_dir: self.file_server.set_roots([(pub_dir,flags_full)])
            
            self.file_server.start()
            self.cache.start()
        
            def signal_handler(signal, frame): 
                raise Error('signal')
            signal.signal(signal.SIGINT , signal_handler)
            signal.signal(signal.SIGTERM, signal_handler)
            if hasattr(signal,"SIGPWR"):
                signal.signal(signal.SIGPWR , signal_handler)

            stopping = 0

            try:
                threadnice_mainloop_fake_start()
                try:
                    while 1:
                        while not select.select([com_sock],[],[],60)[0]:
                            threadnice_mainloop_fake_iteration()

                        connection = com_sock.accept()[0]
                        input  = connection.makefile("rt")
                        output = connection.makefile("wt")
                        command = input.readline()

                        if command == 'quit\n':
                            stopping = 1
                            break
                        elif command[:8] == 'publish ':
                            #if pub_dir: self.file_server.remove_root(pub_dir)
                            pub_dir = command[8:-1]
                            if pub_dir: 
                                self.file_server.set_roots([(pub_dir,flags_full)])
                                output.write(_("Active circle daemon's public directory updated to ")+pub_dir)
                            else:
                                self.file_server.set_roots([ ])
                                output.write(_("Active circle daemon no longer publishing files."))
                        elif command == 'status\n':
                            self.node.lock.acquire()
                            port    = self.node.address[1]
                            n_peers = len(self.node.peers)
                            n_bytes = self.node.network_usage
                            self.node.lock.release()
                    
                            self.file_server.lock.acquire()
                            n_files = len(self.file_server.paths)
                            self.file_server.lock.release()

                            status = _('Circle daemon running on port %d:\n')%port
                            if pub_dir:
                                status = status + _('  %d files published from directory ')%n_files+pub_dir+'\n'
                            else:
                                status = status + _('  No public directory\n')
                            status = status + _('  Connected to %d peers\n  Total network usage %s') % (n_peers,human_size(n_bytes))
                            output.write(status)
                        else:
                            output.write(_('Huh?'))

                        input.close()
                        output.close()
                        connection.close()
                finally:
                    threadnice_mainloop_fake_stop()
            except Error: # signal, eg ^C, SIGTERM
                pass
            except:
                traceback.print_exc()
            
            signal.signal(signal.SIGINT , signal.SIG_IGN)
            signal.signal(signal.SIGTERM, signal.SIG_IGN)
            if hasattr(signal,"SIGPWR"):
                signal.signal(signal.SIGPWR , signal.SIG_IGN)

            self.cache.stop()
            self.file_server.stop()
            self.node.stop()
            
            com_sock.close()
            os.unlink(com_sock_file)
            
            if stopping:
                output.write(_('Circle daemon stopped.'))
                input.close()
                output.close()
                connection.close()

            self.stop()
        else:
            print _("Circle daemon started.")
        
    def run(self, proxy_mode='auto', no_connect=0):
        if no_connect:
            proxy_mode = 'never'

        proxy_host = None
        proxy_password = None

        self.open_window()

        if proxy_mode == 'always' or \
             (proxy_mode == 'auto' and need_proxy()):
            #if sys.platform == 'win32':
            #  Error('The proxy for Masquerading networks does not currently work under Windows, sorry.').show(None,threadnice_mainquit)
                #threadnice_mainloop()
                #return
        
            self.continue_running = 0
            proxy_host, proxy_password = self.run_login_screen(proxy_mode)
        else:
            self.continue_running = 1

        while 1:
            if not self.continue_running:
                break

            self.continue_running = 0
            self.run_main(proxy_host,proxy_password,no_connect)

            if not self.continue_running:
                break
            
            self.continue_running = 0

            proxy_host, proxy_password = self.run_login_screen(proxy_mode)

        self.window.destroy()
    
    def run_main(self, proxy_host,proxy_password, no_connect=0):
        #if proxy_host == None:
        #  try:
#       check_masquerading()
#      except Error, error:
#       error.show(None,mainquit)
#       mainloop()
#       return
    
        sys.path[:0] = [ os.getcwd() ]
        absolutize_path()
        if sys.platform <> 'win32':
            os.chdir(os.environ['HOME'])
        else:
            os.chdir(os.path.dirname(os.path.abspath(sys.argv[0])))
        
        status_setter, progress_setter = self.become_message_window(_("Connecting to the Circle network..."))
        while events_pending():
            mainiteration()
        
        self.node = Node()
        self.file_server = File_server(self.node)
        self.name_server = Name_server(self,self.node,self.window)
        self.chat = Chat_gtk(self)
        self.gossip = Gossip(self)
        self.channels = Channels(self)
        self.cache = Cache(self.node)
        self.download_manager = Download_manager(self)
        
        self.quit_message = None
    
        self.start()

        try:
            self.node.start(proxy_host,proxy_password,self.window,no_connect)
        except Error, error:
            error.show(self.window)
            self.stop()
            self.continue_running = 1
            return

        if not no_connect:
            def waiter(node,n=[0]):
                n[0] = n[0] + 1
                if node.is_connected() or n[0] > 40:
                    threadnice_mainquit()
                    return FALSE
                return TRUE
            schedule_gtk(250,waiter,self.node)
            threadnice_mainloop()
     
        self.set_roots()
        
        try:
            self.name_server.set_name(self.config['name'],self.config['human-name'],self.config['description'])
        except Error:
            sys.exc_info()[1].show(_self.window)

        self.name_server.start(lambda str, self=self: do_gtk(self.chat.show,'\n'+str+'\n\n'))
        
        self.file_server.start()
        self.gossip.start()
        self.channels.start()
        self.cache.start()

        self.become_main_window()
        while events_pending():
            mainiteration()
        
        self.chat.start()
        self.download_manager.start()

        self.start_plug_ins()

        schedule_gtk(5000,self.update)
        #timeout_tag = timeout_add(5000,self.update)

        self.node.lock.acquire()
        any = (len(self.node.peers) <> 0)
        self.node.lock.release()

        if not any and not no_connect:
            Error(_("Couldn't find any peers.\n\n"+
                        "You need to specify the location of at least one peer,\n"+
                        "using the \"Network/Connect to peer...\" menu item.")).show(self.window)

        def signal_handler(signal, frame): 
            print _("Received signal %d, exiting.") % signal
            threadnice_mainquit()
            
        signal.signal(signal.SIGINT , signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        if hasattr(signal,"SIGPWR"):
            signal.signal(signal.SIGPWR , signal_handler)

        threadnice_mainloop()

        signal.signal(signal.SIGINT , signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        if hasattr(signal,"SIGPWR"):
            signal.signal(signal.SIGPWR , signal.SIG_DFL)
        
        self.download_manager.stop()

        close_all_windows()
        while events_pending():
            mainiteration()

        unschedule_gtk(self.update)
        #timeout_remove(timeout_tag)

        #self.running = 0
        self.stop()

        set_config("configuration", self.config)

        #print "Offloading links stored in this node. This may take a little while."
        #print "Please do not interrupt, as doing so will degrade the network."
        str = _("Passing on this peer's obligations.\n\n")
        end_str = '\n\n\n\n'+random.choice(gratuitous_plugs)
        #status = Status()
        #status.show('Shutdown',str,icon="neko.xpm")
        status, progress = self.become_message_window(str+end_str)

        self.node.deactivate_hashtable()
        
        initial_links = len(self.node.links)
        def link_monitor(node,progress,initial_links):
            value = float(initial_links-len(node.links))/initial_links
            progress(value)
            return value < 1.0

        if initial_links:
            schedule_gtk(250, link_monitor, self.node, progress, initial_links)
        
        def shutdown_task(self, status=status,str=str,end_str=end_str):
            self.node.save_seed()

            do_gtk(status,str+_('Stopping services...')+end_str)
        
            def stopper(self,func,*param):
                apply(func,param)
                #print func, param
            stoppers = [
                ("Plug ins",Task(stopper,self,self.stop_plug_ins),None),
                ("Name server",Task(stopper,self,self.name_server.stop,self.quit_message),self.name_server),
                ("File server",Task(stopper,self,self.file_server.stop),self.file_server),
                ("Chat",Task(stopper,self,self.chat.stop),self.chat),
                ("Gossip",Task(stopper,self,self.gossip.stop),self.gossip),
                ("Channels",Task(stopper,self,self.channels.stop),self.channels),
                ("Cache",Task(stopper,self,self.cache.stop),self.cache)
            ]

            for item in stoppers: item[1].start()

            started = time.time()
            for item in stoppers: 
                #wait = 120 - (time.time()-started)
                #if wait > 0.0:
                #  item[1].join(wait)
                while item[1].isAlive() and time.time() < started+240:
                    time.sleep(0.05)

                if item[1].isAlive():
                    sys.stderr.write(item[0] + _(" appears to be deadlocked.\n"))
                    if item[2]:
                        try:
                            sys.stderr.write('Lock: '+`item[2].lock.__dict__`+'\n')
                            for task in item[2].tasks:
                                sys.stderr.write('Task: '+`task.function`+`task.parameters`+'\n')
                        except:
                            pass
                    sys.stderr.write(_("\nIf you can replicate this bug, please report the conditions in which it occurs.\n\n"))

            #do_gtk(_status.message,_str+'Stopping plug-ins...')
            #self.stop_plug_ins()
            #do_gtk(_status.message,_str+'Stopping name server...')
            #self.name_server.stop(self.quit_message)
            #do_gtk(_status.message,_str+'Stopping file server...')
            #self.file_server.stop()
            #do_gtk(_status.message,_str+'Stopping chat server...')
            #self.chat.stop()
            #do_gtk(_status.message,_str+'Stopping gossip server...')
            #self.gossip.stop()

            do_gtk(status,str+_('Stopping peer...')+end_str)
            self.node.stop()
            do_gtk(threadnice_mainquit)

        Task(shutdown_task,self).start()
        threadnice_mainloop()
        #status.close()
        #while events_pending():
        #  mainiteration()

        #shutdown_all_task_managers()

    def run_login_screen(self, proxy_mode):
        vbox = GtkVBox(FALSE, 10)

        label = GtkLabel("The Circle")
        vbox.pack_start(label, TRUE,TRUE,0)

        if proxy_mode in ['always','auto']:
            #sometimes hidden proxy dialog
            proxy_box = GtkVBox(FALSE, 10)
            vbox.pack_start(proxy_box, FALSE,FALSE,0)
            
            if sys.platform == 'win32':
                heading = Helpful_label_widget(_("PuTTY session name:"),
                    _("You appear to be using a masquerading network.\n\n"+\
                    "Circle needs to run a proxy on your firewall in order to work "+\
                    "correctly.\n\n"+\
                    "Enter the name of the PuTTY session for your firewall in the box below.\n\n"+\
                    "For detailed instructions, press the \"Help\" button."))
            else:
                heading = Helpful_label_widget(_("Machine to run proxy on:"),
                    _("You appear to be using a masquerading network.\n\n"+\
                    "Circle needs to run a proxy on your firewall in order to work "+\
                    "correctly. For this you need to have installed ssh and python "+\
                    "on your firewall.\n\n"+
                    "Give the hostname or username@hostname (as in ssh) for your firewall "+\
                    'in the box below then click "log in" to start the proxy.'))
                
            proxy_box.pack_start(heading.widget, FALSE,FALSE,0)

            entry = GtkEntry()
            entry.set_text(get_config('proxy',''))
            proxy_box.pack_start(entry, FALSE,FALSE,0)

            if sys.platform == 'win32':
                label = GtkLabel(_('Password:'))
                label.set_alignment(0,0)
                proxy_box.pack_start(label, FALSE,FALSE,0)

                password_entry = GtkEntry()
                password_entry.set_visibility(0)
                proxy_box.pack_start(password_entry, FALSE,FALSE,0)

            vbox.pack_start(GtkLabel(''), TRUE,TRUE,0)

            # Check for masquerading
            masquerading = [1]

            def on_timeout(masquerading=masquerading,proxy_box=proxy_box):
                need = need_proxy()
                if need and not masquerading[0]:
                    masquerading[0] = 1
                    proxy_box.show_all()
                elif not need and masquerading[0]:
                    masquerading[0] = 0
                    proxy_box.hide()

                return TRUE

            if proxy_mode == 'auto':
                schedule_gtk(1000,on_timeout)
     
        hbox = GtkHBox(FALSE, 0)
        vbox.pack_start(hbox, FALSE,FALSE,5)

        login_button = GtkButton(_("Log in"))
        def on_clicked(button, self=self):
            self.continue_running = 1
            threadnice_mainquit()
        login_button.connect("clicked",on_clicked)
        login_button.set_flags(CAN_DEFAULT)
        hbox.pack_start(login_button, TRUE,TRUE,5)
        
        button = GtkButton(_("Exit"))
        button.connect("clicked",threadnice_mainquit)
        button.set_flags(CAN_DEFAULT)
        hbox.pack_start(button, TRUE,TRUE,5)

        button = GtkButton(_("Help"))
        def on_clicked(button, self=self):
            try:
                browse_file("getting_started.html","#Preliminaries")
            except:
                error.show(self.window)
        button.connect("clicked",on_clicked)
        button.set_flags(CAN_DEFAULT)
        hbox.pack_start(button, TRUE,TRUE,5)

        self.set_window_child(vbox, 1, 10)

        login_button.grab_default()

        if proxy_mode == 'auto':
            on_timeout()

        threadnice_mainloop()

        if proxy_mode in ['always','auto']:
            unschedule_gtk(on_timeout)
            #timeout_remove(timeout_tag)

            if masquerading[0] and self.continue_running:
                text = string.strip(entry.get_text()) 

                if text:
                    set_config('proxy',text)
                    if sys.platform == 'win32':
                        return text, password_entry.get_text()
                    else:
                        return text, None
                else:
                    Error(_("To operate correctly, Circle needs to start a proxy on your firewall.\n\n"+\
                                "Please give the hostname or address of your firewall.")).show(self.window)
                    return self.run_login_screen(proxy_mode)

        return None, None

    def run_text(self, no_connect=0):
        if not no_connect and need_proxy():
            print _("You appear to be using a masquerading network.")
            print _("This isn't currently supported in the text mode version, sorry.")
            return

        self.node = Node()
        
        #self.file_server = File_server(self.node)

        self.name_server = Name_server(self,self.node)
        self.cache = Cache(self.node)
        self.chat = Chat_text(self)
        #self.gossip = Gossip(self)
        self.gossip = None
        self.channels = Channels(self)

        try:
            self.name_server.set_name(self.config['name'],self.config['human-name'],self.config['description'])
        except Error, error:
            print error.message
            return

        self.start()
    
        self.node.start(no_connect=no_connect)

        if not no_connect:
            print _("Connecting to the Circle network...")
            for i in range(40):
                if self.node.is_connected():
                    break
                time.sleep(0.25)

            if not self.node.is_connected():
                # TODO: add greet peer option
                print _("Couldn't find any peers, exiting.")
                return
                
        def signal_handler(signal, frame): 
            raise Error(_('Signal %d received, exiting.') % signal)
        #def sigint_handler(signal, frame):
        #  raise Error('sigint')
        
        signal.signal(signal.SIGINT , signal_handler)
        signal.signal(signal.SIGTERM, signal_handler)
        if hasattr(signal,"SIGPWR"):
            signal.signal(signal.SIGPWR , signal_handler)
        
        self.quit_message = None
        
        print _('Type /help for instructions.')

        def monitor(str, self=self): 
            self.chat.show(str+'\n')
        self.name_server.start(monitor)
        #self.file_server.start()
        self.cache.start()
        #self.gossip.start()
        self.channels.start()
        self.chat.start()
        self.start_plug_ins()

        self.chat.do('/who')

        try:
            self.chat.text_mainloop()
        except Error, error:
            self.chat.show('\n\n'+error.message+'\n\n')
        
        print _("Stopping services...")
        self.chat.stop()
        self.node.deactivate_hashtable()

        self.stop_plug_ins()
        self.cache.stop()
        self.channels.stop()
        self.name_server.stop(self.quit_message)
        #self.gossip.stop()
        #self.file_server.stop()

        print _("Stopping node...")
        self.node.stop()

        self.stop()

    def shutdown(self, message=None, continue_running=0):
        self.continue_running = continue_running
        self.quit_message = message
        threadnice_mainquit()

    def config_dialog(self):
        if self.config_dialog_open:
            return
            
        self.config_dialog_open = 1

        window = GtkWindow()
        window.set_transient_for(self.window)
        window.set_title(_("Configuration"))

        def closed(_b, _self=self):
            _self.config_dialog_open = 0
        window.connect("destroy", closed)
        window.set_border_width(10)
        window.set_default_size(400,10)
        window.set_policy(FALSE,FALSE,TRUE)

        vbox = GtkVBox(FALSE, 10)
        window.add(vbox)

        notebook = GtkNotebook()
        vbox.pack_start(notebook, TRUE,TRUE,0)

        page_vbox = GtkVBox(FALSE,10)
        page_vbox.set_border_width(10)
        notebook.append_page(page_vbox, GtkLabel(_('Identity')))

        label = Helpful_label_widget(_('Username:'),\
            _('A short distinctive name that people can refer to you '+\
            'by. For example your initials or the name of your favourite '+\
            'movie character.\n\n'+\
            'Examples: pfh, jsmith, sonic'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        self.name_entry = GtkEntry()
        self.name_entry.set_text(self.config['name'])
        page_vbox.pack_start(self.name_entry, FALSE,FALSE,0)

        label = Helpful_label_widget(_('Real name:'),\
            _('Your actual real life name.\n\n' +\
            'Example: John Smith'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        self.human_name_entry = GtkEntry()
        self.human_name_entry.set_text(self.config['human-name'])
        page_vbox.pack_start(self.human_name_entry, FALSE,FALSE,0)

        label = Helpful_label_widget(_('About yourself:'),\
            _('Some information about yourself, such as who you are, '+\
            'what you do. Also maybe the URL of your homepage.\n\n' +\
            'Example: I am a chicken wrangler in the outer mongolian highlands. '+\
            'I like Buffy. Visit my homepage at http://wranglers.mo/~bongle/'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        self.description_entry = GtkText()
        self.description_entry.set_point(0)
        self.description_entry.insert_defaults(self.config['description'])
        self.description_entry.set_editable(TRUE)
        self.description_entry.set_word_wrap(TRUE)

        scrolly = GtkScrolledWindow()
        scrolly.set_policy(POLICY_NEVER,POLICY_AUTOMATIC)
        scrolly.set_usize(300,100)
        scrolly.add(self.description_entry)

        page_vbox.pack_start(scrolly, FALSE,FALSE,0)

        label = Helpful_label_widget(_('Signature:'),\
            _('This is a graphical representation of your public key. '+\
            'People can use it to check that you are really you and not an imposter '+\
            'when they search for you.\n\n'+\
            'If you, say, put this on your web-page, people will be able to confirm that '+\
            'they are talking to the person who made that web-page on circle.\n\n'+\
            'Click on the picture to generate a postscript version for printing or putting on teeshirts (Linux and BSD only).\n\n' +\
            'Note: you would only bother doing this if you are a security nut.'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        signature = Signature_widget(self.name_server.public_key_name)
        page_vbox.pack_start(signature.widget,FALSE,FALSE,0)

        page_vbox = GtkVBox(FALSE,10)
        page_vbox.set_border_width(10)
        notebook.append_page(page_vbox, GtkLabel('Directories'))

        label = Helpful_label_widget(_('Public directory:'),\
            _('A directory containing files you want other people to be able to '+\
            'download. Specify the directory relative to your home directory.\n\n'+\
            'If you want to make several directories public, separate them by commas.\n\n'+\
            'Example: public_dir, /mnt/hd2/mp3'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        self.root_entry = File_entry_widget(_('Public directory'),_('Please select your public directory.'),self.config['root'],1,window)
        page_vbox.pack_start(self.root_entry.widget, FALSE,FALSE,0)

        label = Helpful_label_widget(_('Download directory (also public):'),\
            _('Directory to store downloaded files in. ' +\
            'This directory is also public.\n\n'+\
            'Examples: download, public_dir/download, /mnt/hd2/mp3/downloads'))
        page_vbox.pack_start(label.widget,FALSE,FALSE,0)

        self.download_entry = File_entry_widget(_('Download directory'),_('Please select your download directory.\n\nNote: This directory is also public.'),self.config['download'],1,window)
        page_vbox.pack_start(self.download_entry.widget, FALSE,FALSE,0)

        if sys.platform == 'win32' or \
             not os.path.isdir('/var/cache/apt/archives'):
            self.publish_apt_cache_checkbox = None
        else:
            self.publish_apt_cache_checkbox = GtkCheckButton(_('Make apt cache public'))
            self.publish_apt_cache_checkbox.set_active(self.config['publish_apt_cache'])
            enclosure = Helpful_label_widget(self.publish_apt_cache_checkbox,\
                _('Make /var/cache/apt/archives public.\n\n'+\
                    'This will let people use your cached Debian packages, if they use '+\
                    'the Circle apt methods, reducing the load on official Debian servers.'))
            page_vbox.pack_start(enclosure.widget, FALSE,FALSE,10)

        hbox = GtkHBox(FALSE,5)
        vbox.pack_end(hbox, FALSE,FALSE,0)

        button = GtkButton(_("Cancel"))
        button.connect("clicked",lambda _b, _window=window: _window.destroy()) 
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)
        
        button = GtkButton(_("OK"))

        def ok_button(_b, _self=self,_window=window):
            #if _self.config['root']     <> '': _self.file_server.remove_root(_self.config['root'])
            #if _self.config['download'] <> '': _self.file_server.remove_root(_self.config['download'])
            #_self.remove_roots()
            
            _self.config['name']        = string.strip(_self.name_entry.get_text())
            _self.config['human-name']  = string.strip(_self.human_name_entry.get_text())
            _self.config['description'] = string.strip(_self.description_entry.get_chars(0,-1))
            _self.config['download']    = string.strip(_self.download_entry.get_text())
            _self.config['root']        = string.strip(_self.root_entry.get_text())

            if _self.publish_apt_cache_checkbox:
                _self.config['publish_apt_cache'] = _self.publish_apt_cache_checkbox.get_active()

            if _self.config['download'][:2] == '~/': _self.config['download'] = _self.config['download'][2:]
            if _self.config['root'][:2]     == '~/': _self.config['root']     = _self.config['root'][2:]
        
            #set_config("configuration", _self.config)
            
            #if _self.config['root']     <> '': _self.file_server.add_root(_self.config['root'])
            #if _self.config['download'] <> '': _self.file_server.add_root(_self.config['download'])
            _self.set_roots()

            #try:
                #_self.name_server.set_name(_self.config['name'],_self.config['human-name'],_self.config['description'])
            #except Error:
            #  sys.exc_info()[1].show(_self.window)
            _self.sync_config()

            _window.destroy()
            
        button.connect("clicked",ok_button) 
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)

        show_window(window, 'configure')

        button.grab_default()

    def update(self):
        self.file_server.lock.acquire()

        #n = len(self.file_server.paths)
        #if not n:
        #  #str = ''
        #  self.info_files_button.hide()
        #else:
        #  #str = '%d files published. ' % n
        #  self.info_files_button.show_all()
        #  self.info_files_label.set_text(_('%d files published.')%n)

        #str = 'Network usage so far: %.1fMB.' % (self.node.network_usage/float(1<<20))
        #str = _('Network usage so far: ') + human_size(self.node.network_usage)
            
        #self.info_net_label.set_text(str)

        self.file_server.lock.release()

        self.drawing.queue_draw()
        
        return TRUE

    def open_window(self):
        self.window_pos = get_config("window",None)
        if not self.window_pos:
            self.window_pos = (500,500,1,1)

        self.window = GtkWindow(WINDOW_TOPLEVEL)
        #self.window.set_wmclass('main','the_circle')
        self.window.set_default_size(self.window_pos[0],self.window_pos[1])
        self.set_title()
    
        self.mainquit_on_delete = 0

        def on_delete(window, event, self=self):
            if self.mainquit_on_delete:
                threadnice_mainquit()

            return TRUE # Don't close window
        self.window.connect("delete_event",on_delete)

        # Text sometimes fails to draw correctly
        if sys.platform == 'win32':
            def on_focus(window, event):
                window.queue_draw()
            self.window.connect("focus-in-event",on_focus)

        #self.window.show()
        show_window(self.window,'main',0)

        self.current_window_child = None

    def set_title(self, title = None):
        if not title:
            title = _("The Circle ")+__init__.__version__
        self.window.set_title(title)

    def set_window_child(self, child, allow_close, border=0):
        if self.current_window_child:
            self.window.remove(self.current_window_child)

        self.mainquit_on_delete = allow_close
        
        self.window.set_border_width(border)
            
        self.current_window_child = child
        self.window.add(child)
        child.show_all()

    def become_message_window(self, message):
        hbox = GtkHBox(FALSE,0)

        vbox = GtkVBox(FALSE,0)
        hbox.pack_start(vbox, FALSE,FALSE,0)

        drawing = GtkDrawingArea()
        vbox.pack_start(drawing, TRUE,TRUE,0)
        def expose(widget,event):
            widget.draw_rectangle(widget.get_style().black_gc,1, event.area[0],event.area[1],event.area[2],event.area[3])
        drawing.connect("expose_event",expose)

        try:
            pixmap = GtkPixmap(self.window,find_file("neko.xpm"))
            vbox.pack_start(pixmap, FALSE,FALSE,0)
        except Error:
            pass
        
        #win32 crashes on set_line_wrap
        #label = GtkLabel(message)
        #label.set_line_wrap(TRUE)
        #hbox.pack_start(label,TRUE,TRUE,0)

        vbox_inner = GtkVBox(FALSE, 0)
        hbox.pack_start(vbox_inner,TRUE,TRUE,10)

        align = GtkAlignment(0.5,0.5,0,0)
        vbox_inner.pack_start(align,TRUE,TRUE,0)

        text = Text_widget(300)
        text.write(message)
        align.add(text.widget)

        progress = GtkProgressBar('')
        vbox_inner.pack_start(progress, FALSE,FALSE,10) 

        self.set_window_child(hbox, 0)

        progress.hide()

        def set_text(message, text=text):
            text.clear()
            text.write(message)
            text.widget.queue_draw()

        def set_progress(amount, progress=progress):
            progress.update(amount)
            progress.show()

        #return lambda message, label=label: label.set_text(message)
        return (set_text, set_progress)

    def add_menu_item(self,menu_name,name,action,*args):
        """Add a menu item to a menu. If the menu does not exist it will
             also be created.

             If the menu item has an equivalen chat command, add that command
             to the name preceded by a colon, eg 'Log out:/logout'
        
             Action will be called with the menu item object, followed by
             *args """

        if not self.menus.has_key(menu_name):
            menu = GtkMenu()
            menu_item = GtkMenuItem(menu_name)
            menu_item.set_submenu(menu)
            self.menu_bar_widget.insert(menu_item,len(self.menus)-1) # Don't add after Help menu
            menu_item.show_all()
            self.menus[menu_name] = menu

        name_parts = string.split(name,':')

        if len(name_parts) == 1:
            menu_item = GtkMenuItem(name)
        else:
            menu_item = GtkMenuItem()

            hbox = GtkHBox(FALSE, 20)
            menu_item.add(hbox)

            hbox.pack_start(GtkLabel(name_parts[0]), FALSE,FALSE,0)
            hbox.pack_end(GtkLabel(name_parts[1]), FALSE,FALSE,0)
        
        apply(menu_item.connect,("activate",action)+args)
        menu_item.show()
        self.menus[menu_name].append(menu_item)

        return menu_item

    def add_menu_separator(self,menu_name):
        """Add a separator to a menu."""
        menu_item = GtkMenuItem()
        menu_item.show()
        self.menus[menu_name].append(menu_item)

    def become_main_window(self):
        main_vbox = GtkVBox(FALSE,5)

        top_hbox = GtkHBox(FALSE,0)
        main_vbox.pack_start(top_hbox, FALSE,FALSE,0)

        vbox = GtkVBox(FALSE,5)
        top_hbox.pack_start(vbox, TRUE,TRUE,0)

        menu_bar = GtkMenuBar()
        menu_bar.set_shadow_type(SHADOW_NONE)
        #main_vbox.pack_start(menu_bar, FALSE,FALSE,0)
        vbox.pack_start(menu_bar, FALSE,FALSE,0)

        menu = GtkMenu()
        menu_item = GtkMenuItem(_('Help'))
        menu_item.set_submenu(menu)
        menu_bar.append(menu_item)

        self.menu_bar_widget = menu_bar
        self.menus = { _('Help'): menu }

        self.add_menu_item(_('Network'),_('Configure...'),self.menu_configure)
        self.add_menu_separator(_('Network'))
        self.add_menu_item(_('Network'),_('Browse current peers'),self.menu_browse_peers)
        self.add_menu_item(_('Network'),_('Connect to peer...'),self.menu_greet_peer)
        self.add_menu_item(_('Network'),_('Network traffic monitor'),lambda m,self=self: node_monitor.monitor_node(self.node,self.file_server))
        self.add_menu_separator(_('Network'))
        self.add_menu_item(_('Network'),_('Browse locally published files'),lambda m,self=self: search_browse_files(self.node.address, self.node,self))
        self.add_menu_separator(_('Network'))
        self.add_menu_item(_('Network'),_('Log out')+':/logout',self.menu_logout)
        self.add_menu_item(_('Network'),_('Exit')+':/exit',self.menu_exit)

        self.add_menu_item(_('Chat'),_('Browse people'),self.menu_browse_all)
        self.add_menu_item(_('Chat'),_('Browse channels'),self.menu_browse_channels)
        self.add_menu_item(_('Chat'),_('Create new channel...'),self.menu_create_channel)
        self.add_menu_separator(_('Chat'))
        self.add_menu_item(_('Chat'),_('Toggle quiet')+':/quiet',lambda m,self=self: self.chat.do('/quiet'))
        self.add_menu_item(_('Chat'),_('Read unread messages')+':/read',lambda m,self=self: self.chat.do('/read'))

        self.add_menu_item(_('Gossip'),_('Post gossip...')+':/post',lambda m,self=self: self.gossip.prompt_for_wodge())
        self.add_menu_item(_('Gossip'),_('Show hot gossip')+':/gossip',lambda m,self=self: self.gossip.update_and_show_gossip())

        self.add_menu_item(_('Help'),_('Getting started'),self.menu_getting_started)
        self.add_menu_item(_('Help'),_('Chat commands')+':/help',self.menu_chat_commands)
        self.add_menu_item(_('Help'),_('Running Circle as a daemon'),self.menu_daemon_howto)
        self.add_menu_item(_('Help'),_('Frequently Asked Questions'),self.menu_faq)
        self.add_menu_separator(_('Help'))
        self.add_menu_item(_('Help'),_('Developer documentation'),self.menu_pydoc)
        self.add_menu_separator(_('Help'))
        self.add_menu_item(_('Help'),_('Homepage'),self.menu_homepage)
        self.add_menu_item(_('Help'),_('About'),self.menu_about)

        #self.info_label = GtkLabel('')
        #self.info_label.set_alignment(0,0)
        #vbox.pack_end(self.info_label, FALSE,FALSE,0)

        #info_hbox = GtkHBox(FALSE,5)
        #vbox.pack_end(info_hbox, FALSE,FALSE,0)

        #self.info_files_button = GtkButton()
        #self.info_files_button.set_relief(RELIEF_NONE)
        #self.info_files_button.connect('clicked',lambda b, self=self: search_browse_files(self.node.address, self.node,self))
        #info_hbox.pack_start(self.info_files_button, FALSE,FALSE,0)
        #self.info_files_label = GtkLabel('')
        #self.info_files_button.add(self.info_files_label)
        #self.info_net_label = GtkLabel('')
        #info_hbox.pack_start(self.info_net_label, FALSE,FALSE,0)

        padding = GtkLabel('')
        top_hbox.pack_start(padding, FALSE,FALSE, 5)

        button = GtkButton()
        button.connect("clicked",self.menu_browse_peers)
        button.set_relief(RELIEF_NONE)
        top_hbox.pack_start(button, FALSE,FALSE,0)
        tooltips.set_tip(button,_("current peers"),"")

        if sys.platform == 'win32':
            button.set_sensitive(FALSE)

        #self.drawing = GtkDrawingArea()
        self.drawing = GtkPixmap(self.window,find_file("diagram-bg.xpm"))
        #self.drawing.set_alignment(0,0)
        #self.drawing.size(drawing_size+1+10,drawing_size+1)
        #self.drawing.size(drawing_size+2,drawing_size+2)
        def expose(widget,event, _self=self):
            alloc = widget.get_allocation()
            line_scale = drawing_size*15/32
            line_shift_x = alloc[0]+(alloc[2]+1)/2
            line_shift_y = alloc[1]+(alloc[3]+1)/2
            lines = [ ]
            
            _self.node.lock.acquire()

            me = hash.float_hash(_self.node.name)*2.0*math.pi
            for peer in _self.node.peers.values():
                it = hash.float_hash(peer.name)*2.0*math.pi
                it_p = complex(math.cos(it),math.sin(it))
                me_p = complex(math.cos(me),math.sin(me))

                if not _self.node.hashtable_running:
                    lines.append(
                        [ (it_p.real*line_scale+line_shift_x,it_p.imag*line_scale+line_shift_y),
                            (it_p.real*line_scale*0.7+line_shift_x,it_p.imag*line_scale*0.7+line_shift_y) ])
                    continue

                chord_len = abs(it_p - me_p)
                if chord_len < 0.001:
                    continue

                # Simpler code for drawing between almost-opposite points.
                if abs(it_p + me_p) * line_scale < 2 * 3.0:
                    lines.append(
                        [ (me_p.real*line_scale+line_shift_x,me_p.imag*line_scale+line_shift_y),
                            (it_p.real*line_scale+line_shift_x,it_p.imag*line_scale+line_shift_y) ])
                    continue


                # todo: ought to make n_segments a function of tot_rot: adding segments
                # beyond a certain number probably actually looks worse (depending on
                # quality of line-drawing algorithm).  Perhaps have a target value of
                # rot.imag * (segment length in pixels), i.e. * line_scale * abs(delta).

                n_bends = 8
                n_segments = n_bends + 1
                list = [ ]
                point = me_p
                tot_rot = (it_p / -me_p)
                rot = tot_rot ** (1.0/n_bends)

                # Calculate delta, which is the segment vector.
                # First suppose that we used a segment length of 1, and see how long an
                # "arc" that would result in.  This is abs(sum_{i=0}^{nbends} rot^i).
                # Power series simplification: sum_{i=0}^{nbends} rot^i ==
                # (rot^{nbends+1} - 1) / (rot - 1).
                # To calculate rot^{nbends+1}, we use rot^{nbends} * rot^{1},
                # i.e. tot_rot * rot.
                # Having thus calculated the arc length for segments of length 1,
                # we compare with our desired arc length (i.e. chord_len);
                # let delta length = desired "arc" length / abs(that sum), or rather
                # * abs(reciprocal of that sum) given that it's a fraction.
                # Finally, multiply this length by the desired initial direction of -me_p.

                delta = (chord_len * abs((rot - 1)
                                                                 / ((tot_rot * rot) - 1))
                                 * -me_p)

                for i in range(n_segments + 1):
                    list.append((point.real*line_scale+line_shift_x,point.imag*line_scale+line_shift_y))
                    point = point + delta
                    delta = delta * rot
                
                lines.append(list)

            #p = int(drawing_size * _self.node.call_p_success)
            #widget.draw_rectangle(widget.get_style().black_gc,1, drawing_size+2,drawing_size-p,9,p)
            #widget.draw_rectangle(widget.get_style().white_gc,1, drawing_size+2,0,9,drawing_size-p)
            _self.node.lock.release()
            
            style = widget.get_style()
            window = widget.get_window()

            #draw_arc(window,style.light_gc[0],TRUE,0,0,drawing_size,drawing_size, 64*0,64*180)
            #draw_arc(window,style.light_gc[0],TRUE,0,0,drawing_size,drawing_size, 64*180,64*360)

            for list in lines:
                draw_lines(window,style.light_gc[0],list)
            
            #for pair in [(0,0),(0,1),(1,0),(1,1)]:
                #draw_arc(window,style.dark_gc[0],FALSE,pair[0],pair[1],drawing_size,drawing_size, 64*0,64*180)
                #draw_arc(window,style.dark_gc[0],FALSE,pair[0],pair[1],drawing_size,drawing_size, 64*180,64*360)
                    
        #self.drawing.connect("expose_event",expose)
        self.drawing.connect_after("expose-event",expose)
        self.drawing.connect_after("draw",expose)
        button.add(self.drawing)

        #hbox = GtkHBox(FALSE,5)
        search_box = GtkVBox(FALSE,0)
        vbox.pack_start(search_box, TRUE,FALSE,0)

        #option_menu = GtkOptionMenu()
        #hbox.pack_start(option_menu, TRUE,TRUE,0)
        #menu = GtkMenu()
        #option_menu.set_menu(menu)

        #def select(_i, id, _self=self):
        #  _self.search_mode = id
        #item = GtkMenuItem('File')
        #item.connect("activate",select,0)
        #menu.append(item)
        #item = GtkMenuItem('Person')
        #item.connect("activate",select,1)
        #menu.append(item)
        
        #option_menu.set_history(0)
        #self.search_mode = 0
        
        search_hbox = GtkHBox(FALSE,0)
        search_box.pack_start(search_hbox, FALSE,FALSE,0)

        search_hbox.pack_start(GtkLabel(_('Search: ')), FALSE,FALSE,5)

        self.search_entry = GtkEntry()
        def entry_enter(_e, event, _self=self):
            if event.keyval == 65293: # enter keyval
                _self.search(_e,0)
        self.search_entry.connect("key-press-event",entry_enter)
        search_hbox.pack_start(self.search_entry)

        #button = GtkButton(" Search ")
        #button.connect("clicked", self.search)
        #hbox.pack_start(button, FALSE,FALSE,0)

        search_hbox = GtkHBox(FALSE,0)
        search_box.pack_start(search_hbox, FALSE,FALSE,0)

        button = GtkButton(_(" Find person "))
        button.connect("clicked", self.search, 1)
        search_hbox.pack_end(button, FALSE,FALSE,0)

        button = GtkButton(_(" Find file "))
        button.connect("clicked", self.search, 0)
        search_hbox.pack_end(button, FALSE,FALSE,5)

        hbox = GtkHBox(FALSE,0)
        main_vbox.pack_start(hbox, TRUE,TRUE,0)
        
        hbox.pack_start(self.chat.widget, TRUE,TRUE,0)

        vbox = GtkVBox(FALSE,0)
        hbox.pack_start(vbox, FALSE,FALSE,0)

        button1 = GtkButton()
        button1.set_relief(RELIEF_NONE)
        vbox.pack_start(button1, FALSE,FALSE,0)
        arrow1 = GtkArrow(ARROW_UP,SHADOW_IN);
        button1.add(arrow1)
        tooltips.set_tip(button1,_("show/hide menu"),"")
        
        button2 = GtkButton()
        button2.set_relief(RELIEF_NONE)
        vbox.pack_start(button2, FALSE,FALSE,0)
        arrow2 = GtkArrow(ARROW_RIGHT,SHADOW_IN);
        button2.add(arrow2)
        tooltips.set_tip(button2,_("show/hide people"),"")
        
        scroll_bar = GtkVScrollbar(self.chat.widget.get_vadjustment())
        vbox.pack_start(scroll_bar, TRUE,TRUE,0)
        
        userlist_bin = GtkAlignment(0,0,1,1) # GtkBin broken
        userlist_bin.add(self.chat.userlist_widget)
        hbox.pack_start(userlist_bin, FALSE,FALSE,0)

        #def arrow1_click(button, arrow1=arrow1, mf=mf, top_hbox=top_hbox):
        def arrow1_click(button, arrow1=arrow1, menu_bar=menu_bar, top_hbox=top_hbox):
            if menu_bar.flags(VISIBLE):
                menu_bar.hide()
                top_hbox.hide()
                arrow1.set(ARROW_DOWN,SHADOW_IN)
            else:
                menu_bar.show()
                top_hbox.show()
                arrow1.set(ARROW_UP,SHADOW_IN)
        button1.connect("clicked",arrow1_click)

        def arrow2_click(button, arrow2=arrow2,userlist_bin=userlist_bin):
            if userlist_bin.flags(VISIBLE):
                userlist_bin.hide()
                arrow2.set(ARROW_LEFT,SHADOW_IN)
            else:
                userlist_bin.show()
                arrow2.set(ARROW_RIGHT,SHADOW_IN)
        button2.connect("clicked",arrow2_click)
        
        def on_destroy(_w, window=self.window, menu_bar=menu_bar,userlist_bin=userlist_bin):
            allocation = window.get_allocation()
            set_config("window",(allocation[2],allocation[3],menu_bar.flags(VISIBLE),userlist_bin.flags(VISIBLE)))
            #mainquit()
        main_vbox.connect("destroy", on_destroy)

        self.set_window_child(main_vbox, 1)

        if not self.window_pos[2]:
            arrow1_click(button1)
        if len(self.window_pos) > 3 and not self.window_pos[3]:
            arrow2_click(button2)

        self.chat.widget.grab_focus()

    def menu_greet_peer(self, _b):
        window = GtkWindow()
        window.set_border_width(10)        
        window.set_default_size(300,10)
        window.set_title(_('Connect to peer'))

        window.set_transient_for(self.window)
        window.set_policy(FALSE,FALSE,TRUE)

        vbox = GtkVBox(FALSE, 5)
        window.add(vbox)

        label = Helpful_label_widget(_('Address of peer to connect to:                           '),\
            _("If all has gone well, you will not need to use this dialog box at "+\
            "all. You should see a number of arcs in the circle diagram. This "+\
            "indicates that you are connected to the network. If not, you need "+\
            "to give the address of a peer to connect to. Circle will then ask it about "+\
            "other peers and thus connect to the network.\n\n"+\
            "Give the location of the peer you wish to connect to. This can be either "+\
            "a hostname or an IP address. It can be optionally followed by a "+\
            "colon and a port number.\n\n"+
            "Examples: 130.194.67.43, hawthorn.csse.monash.edu.au, "+\
            " hawthorn.csse.monash.edu.au:29610"))
        vbox.pack_start(label.widget, FALSE,FALSE,0)
        
        entry = GtkEntry()
        vbox.pack_start(entry, FALSE,FALSE,0)

        hbox = GtkHBox(FALSE,5)
        vbox.pack_end(hbox, FALSE,FALSE,0)

        button = GtkButton(_("Cancel"))
        button.connect("clicked",\
            lambda _b, _window=window: _window.destroy())
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)

        def greet(_b, _entry=entry, _window=window, _self=self):
            try:
                address = parse_address(_self.node.address, _entry.get_text())
                Task(probe_task,_self.node,address).start()
                _window.destroy()
            except Error:
                sys.exc_info()[1].show(_window)
        
        button = GtkButton(_("Connect"))
        button.connect("clicked",greet)
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)

        show_window(window, 'greet')

        button.grab_default()
        
    def menu_browse_peers(self, _b=None):
        list = [ ]
        self.node.lock.acquire()
        for peer in self.node.peers.values():
            list.append(peer.address)
        self.node.lock.release()
        search_show_sources(list, self.node,self, _('Current peers'))

    def menu_browse_all(self, _b=None):
        try:
            search_for_all_people(self.node,self)
        except Error:
            sys.exc_info()[1].show(self.window)

    def menu_create_channel(self, _b=None):
        window = GtkWindow()
        window.set_border_width(10)        
        window.set_default_size(300,10)
        window.set_title(_('Create a channel'))

        window.set_transient_for(self.window)
        window.set_policy(FALSE,FALSE,TRUE)

        vbox = GtkVBox(FALSE, 5)
        window.add(vbox)

        label = Helpful_label_widget(_('Name of channel to create:     '),\
            _("This should be the name of the channel you wish to create, "+\
            "for example #circle-devel or #trek"))
        vbox.pack_start(label.widget, FALSE,FALSE,0)
        
        entry = GtkEntry()
        vbox.pack_start(entry, FALSE,FALSE,0)

        hbox = GtkHBox(FALSE,5)
        vbox.pack_end(hbox, FALSE,FALSE,0)

        button = GtkButton(_("Cancel"))
        button.connect("clicked",\
            lambda _b, _window=window: _window.destroy())
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)

        def create(_b, _entry=entry, _window=window, _self=self):
            try:
                chnl=_entry.get_text()
                if not chnl:
                    return
                if chnl[0]!='#':
                    chnl='#'+chnl;
                _self.channels.sub(chnl)
                _window.destroy()
            except Error:
                sys.exc_info()[1].show(_window)
        
        button = GtkButton(_("Create"))
        button.connect("clicked",create)
        button.set_flags(CAN_DEFAULT)
        hbox.pack_end(button, TRUE,TRUE,0)

        show_window(window, 'create channel')

        button.grab_default()
        
    def menu_browse_channels(self, _b=None):
        try:
            search_for_all_channels(self.node,self)
        except Error:
            sys.exc_info()[1].show(self.window)
 
    def menu_configure(self, _b=None):
        self.config_dialog()

    def menu_logout(self, _b=None):
        self.continue_running = 1
        threadnice_mainquit()
        
    def menu_exit(self, _b=None):
        threadnice_mainquit()

    def menu_getting_started(self, _b=None):
        try:
            browse_file('getting_started.html')
        except Error, error:
            error.show(self.window)
    
    def menu_chat_commands(self, _b=None):
        try:
            browse_file('chat_commands.html')
        except Error, error:
            error.show(self.window)

    def menu_daemon_howto(self, _b=None):
        try:
            browse_file('daemon_howto.html')
        except Error, error:
            error.show(self.window)

    def menu_faq(self, _b=None):
        try:
            browse_file('technical_faq.html')
        except Error, error:
            error.show(self.window)

    def menu_homepage(self, _b=None):
        try:
            browse_url('http://www.csse.monash.edu.au/~pfh/circle/')
        except Error, error:
            error.show(self.window)
    
    def menu_pydoc(self, _b=None):
        try:
            browse_modules()
        except Error, error:
            error.show(self.window)

    def menu_about(self, _b=None):
        try:
            browse_file('about.html')
        except Error, error:
            error.show(self.window)

    def search(self, _b, mode):
        try:
            text = self.search_entry.get_text()
            text = string.strip(text)

            #maybe_name = circle_urls.text_to_name_if_md5sum(text)
            #if maybe_name != None:
            #  search.search_for_name(maybe_name, _('Files with specified md5sum:'), self.node, self)
            #if text[:12] == 'circle-file:':
            
            try:
                search_for_name(hash.url_to_hash(text),_('Files matching url'), self.node,self) 
            except Error:
                if mode == 1:
                    if text == '':
                        self.menu_browse_all()
                    else:
                        search_for_people(text, self.node,self)
                else:
                    search_for_files(text, self.node,self)

            self.search_entry.set_text('')
        except Error, error:
            error.show(self.window)

    def start_plug_ins(self):
        plug_in_dir = os.path.join(config_dir,'plug-ins')
        if not os.path.isdir(plug_in_dir):
            os.mkdir(plug_in_dir)
        
        names = os.listdir(plug_in_dir)
        self.plug_ins = [ ]
        for name in names:
            if name[-3:] == '.py':
                realname = os.path.join(plug_in_dir,name)
                #file = open(realname,'rt')
                #module = imp.load_source('circle-plug-in-'+name,realname,file)
                module = { 'app' : self, 'node' : self.node }
                execfile(realname,module)
                #file.close()

                #plug_in = module.Plug_in(self)
                #plug_in.start()
                #self.plug_ins.append(plug_in)
                if module.has_key('start'):
                    module['start']()
                self.plug_ins.append(module)
    
    def stop_plug_ins(self):
        for plug_in in self.plug_ins:
            #plug_in.stop()
            if plug_in.has_key('stop'):
                plug_in['stop']()

    def set_roots(self):
        roots = [ ]
        if self.config['root']     <> '': roots.extend(map(string.strip,string.split(self.config['root'],',')))
        if self.config['download'] <> '': roots.append(self.config['download'])
        
        for i in range(len(roots)):
            roots[i] = (roots[i], flags_full)
        
        if self.config['publish_apt_cache'] : roots.append(('/var/cache/apt/archives',flags_deb))

        self.file_server.set_roots(roots)

    #def add_roots(self):
    #  for root in self.roots():
    #    self.file_server.add_root(root)

    #def remove_roots(self):
    #  for root in self.roots():
    #    self.file_server.remove_root(root)
        
    def sync_config(self):
        name = self.config['name']
        name = string.lower(name)
        name = string.replace(name, " ","")
        name = string.replace(name, "-","")
        self.config['name'] = name
    
        self.name_server.set_name(self.config['name'],self.config['human-name'],self.config['description'])

        set_config("configuration", self.config)

